import importlib
import os
from pathlib import Path
from types import ModuleType
from typing import Generator, List, Type, TypeVar

from flupy import walk_files

T = TypeVar("T")


def walk_modules(module: ModuleType) -> Generator[ModuleType, None, None]:
    """Recursively yield python import paths to submodules in *module*

    Example:
        module_iter = walk_modules(alembic_utils_extended)

        for module_path in module_iter:
            print(module_path)

        # alembic_utils_extended.exceptions
        # alembic_utils_extended.on_entity_mixin
        # ...
    """
    top_module = module
    top_path = Path(top_module.__path__[0])
    top_path_absolute = top_path.resolve()

    directories = (
        walk_files(str(top_path_absolute))
        .filter(lambda x: x.endswith(".py"))
        .map(Path)
        .group_by(lambda x: x.parent)
        .map(lambda x: (x[0], x[1].collect()))
    )

    for base_path, files in directories:
        if str(base_path / "__init__.py") in [str(x) for x in files]:
            for module_path in files:
                if "__init__.py" not in str(module_path):

                    # Example: elt.settings
                    module_import_path = str(module_path)[len(str(top_path_absolute)) - len(top_module.__name__) :].replace(
                        os.path.sep, "."
                    )[:-3]

                    module = importlib.import_module(module_import_path)
                    yield module


def collect_instances(module: ModuleType, class_: Type[T]) -> List[T]:
    """Collect all instances of *class_* defined in *module*

    Note: Will import all submodules in *module*. Beware of import side effects
    """

    found: List[T] = []

    for module_ in walk_modules(module):

        for _, variable in module_.__dict__.items():

            if isinstance(variable, class_):
                # Ensure variable is not a subclass
                if variable.__class__ == class_:
                    found.append(variable)
    return found


def collect_subclasses(module: ModuleType, class_: Type[T]) -> List[Type[T]]:
    """Collect all subclasses of *class_* currently imported or defined in *module*

    Note: Will import all submodules in *module*. Beware of import side effects
    """

    found: List[Type[T]] = []

    for module_ in walk_modules(module):

        for _, variable in module_.__dict__.items():
            try:
                if issubclass(variable, class_) and not class_ == variable:
                    found.append(variable)
            except TypeError:
                # argument 2 to issubclass must be a class ....
                pass

    imported: List[Type[T]] = list(class_.__subclasses__())  # type: ignore

    return list(set(found + imported))



from dataclasses import dataclass
from enum import Enum
from typing import Generator, List, Optional, Union

from flupy import flu
from sqlalchemy import text as sql_text
from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import BadInputException
from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import (
    coerce_to_quoted,
    coerce_to_unquoted,
)


class PGGrantTableChoice(str, Enum):
    # Applies at column level
    SELECT = "SELECT"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    REFERENCES = "REFERENCES"
    # Applies at table Level
    DELETE = "DELETE"
    TRUNCATE = "TRUNCATE"
    TRIGGER = "TRIGGER"

    def __str__(self) -> str:
        return str.__str__(self)

    def __repr__(self) -> str:
        return f"'{str.__str__(self)}'"


C = PGGrantTableChoice


@dataclass(frozen=True, eq=True, order=True)
class SchemaTableRole:
    schema: str
    table: str
    role: str
    grant: PGGrantTableChoice
    with_grant_option: str  # 'YES' or 'NO'


@dataclass
class PGGrantTable(ReplaceableEntity):
    """A PostgreSQL Grant Statement compatible with `alembic revision --autogenerate`

    PGGrantTable requires the role name being used to generate migrations to
    match the role name that executes migrations.

    If your system does not meet that requirement, disable them by excluding PGGrantTable
    in `include_object` https://alembic.sqlalchemy.org/en/latest/api/runtime.html#alembic.runtime.environment.EnvironmentContext.configure.params.include_object

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **table** - *str*: The table to grant access to
    * **columns** - *List[str]*: A list of column names on *table* to grant access to
    * **role** - *str*: The role to grant access to
    * **grant** - *Union[Grant, str]*: On of SELECT, INSERT, UPDATE, DELETE, TRUNCATE, REFERENCES, TRIGGER
    * **with_grant_option** - *bool*: Can the role grant access to other roles
    """

    schema: str
    table: str
    columns: List[str]
    role: str
    grant: PGGrantTableChoice
    with_grant_option: bool

    type_ = "grant_table"

    def __init__(
        self,
        schema: str,
        table: str,
        role: str,
        grant: Union[PGGrantTableChoice, str],
        columns: Optional[List[str]] = None,
        with_grant_option=False,
    ):
        self.schema: str = coerce_to_unquoted(schema)
        self.table: str = coerce_to_unquoted(table)
        self.columns: List[str] = sorted(columns) if columns else []
        self.role: str = coerce_to_unquoted(role)
        self.grant: PGGrantTableChoice = PGGrantTableChoice(grant)
        self.with_grant_option: bool = with_grant_option
        self.signature = self.identity

        if PGGrantTableChoice(self.grant) in {C.SELECT, C.INSERT, C.UPDATE, C.REFERENCES}:
            if len(self.columns) == 0:
                raise BadInputException(f"When grant type is {self.grant} a value must be provided for columns")
        else:
            if self.columns:
                raise BadInputException(f"When grant type is {self.grant} a value must not be provided for columns")

    @classmethod
    def from_sql(cls, sql: str) -> "PGGrantTable":
        raise NotImplementedError()

    @property
    def identity(self) -> str:
        """A string that consistently and globally identifies a function"""
        # rows in information_schema.role_column_grants are uniquely identified by
        # the columns listed below + the grantor
        # be cautious when editing
        return f"{self.__class__.__name__}: {self.schema}.{self.table}.{self.role}.{self.grant}"

    @property
    def definition(self) -> str:  # type: ignore
        return str(self)

    def to_variable_name(self) -> str:
        """A deterministic variable name based on PGFunction's contents"""
        schema_name = self.schema.lower()
        table_name = self.table.lower()
        role_name = self.role.lower()
        return f"{schema_name}_{table_name}_{role_name}_{str(self.grant)}".lower()

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__

        return f"""{var_name} = {self}\n"""

    @classmethod
    def from_database(cls, sess: Session, schema: str = "%"):
        # COLUMN LEVEL
        sql = sql_text(
            """
        SELECT
            table_schema as schema,
            table_name,
            grantee as role_name,
            privilege_type as grant_option,
            is_grantable,
            column_name
        FROM
            information_schema.role_column_grants rcg
            -- Cant revoke from superusers so filter out those recs
            join pg_roles pr
                on rcg.grantee = pr.rolname
        WHERE
            not pr.rolsuper
            and grantor = CURRENT_USER
            and table_schema like :schema
            and privilege_type in ('SELECT', 'INSERT', 'UPDATE', 'REFERENCES')
        """
        )

        rows = sess.execute(sql, params={"schema": schema}).fetchall()
        grants = []

        grouped = flu(rows).group_by(lambda x: SchemaTableRole(*x[:5])).map(lambda x: (x[0], x[1].map_item(5).collect())).collect()
        for s_t_r, columns in grouped:
            grant = cls(
                schema=s_t_r.schema,
                table=s_t_r.table,
                role=s_t_r.role,
                grant=s_t_r.grant,
                with_grant_option=s_t_r.with_grant_option == "YES",
                columns=columns,
            )
            grants.append(grant)

        # TABLE LEVEL
        sql = sql_text(
            """
        SELECT
            table_schema as schema_name,
            table_name,
            grantee as role_name,
            privilege_type as grant_option,
            is_grantable
        FROM
            information_schema.role_table_grants rcg
            -- Cant revoke from superusers so filter out those recs
            join pg_roles pr
                on rcg.grantee = pr.rolname
        WHERE
            not pr.rolsuper
            and grantor = CURRENT_USER
            and table_schema like :schema
            and privilege_type in ('DELETE', 'TRUNCATE', 'TRIGGER')
        """
        )

        rows = sess.execute(sql, params={"schema": schema}).fetchall()

        for schema_name, table_name, role_name, grant_option, is_grantable in rows:
            grant = cls(
                schema=schema_name,
                table=table_name,
                role=role_name,
                grant=grant_option,
                with_grant_option=is_grantable == "YES",
            )
            grants.append(grant)
        return grants

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create view" statement"""
        with_grant_option = " WITH GRANT OPTION" if self.with_grant_option else ""
        maybe_columns_clause = f'( {", ".join(self.columns)} )' if self.columns else ""
        yield sql_text(
            f"GRANT {self.grant} {maybe_columns_clause} ON {self.literal_schema}.{coerce_to_quoted(self.table)} TO {coerce_to_quoted(self.role)} {with_grant_option}"
        )

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop view" statement"""
        # cascade has no impact
        yield sql_text(f"REVOKE {self.grant} ON {self.literal_schema}.{coerce_to_quoted(self.table)} FROM {coerce_to_quoted(self.role)}")

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        yield from self.to_sql_statement_drop()
        yield from self.to_sql_statement_create()



import logging
from contextlib import contextmanager
from typing import Generator, List

from sqlalchemy import exc as sqla_exc
from sqlalchemy.orm import Session

from alembic_utils_extended.simulate import simulate_entity

logger = logging.getLogger(__name__)


def solve_resolution_order(sess: Session, entities):
    """Solve for an entity resolution order that increases the probability that
    a migration will suceed if, for example, two new views are created and one
    refers to the other

    This strategy will only solve for simple cases
    """

    resolved = []

    # Resolve the entities with 0 dependencies first (faster)
    logger.info("Resolving entities with no dependencies")
    for entity in entities:
        try:
            with simulate_entity(sess, entity):
                resolved.append(entity)
        except (sqla_exc.ProgrammingError, sqla_exc.InternalError) as exc:
            continue

    # Resolve entities with possible dependencies
    logger.info("Resolving entities with dependencies. This may take a minute")
    for _ in range(len(entities)):
        n_resolved = len(resolved)

        for entity in entities:
            if entity in resolved:
                continue

            try:
                with simulate_entity(sess, entity, dependencies=resolved):
                    resolved.append(entity)
            except (sqla_exc.ProgrammingError, sqla_exc.InternalError):
                continue

        if len(resolved) == n_resolved:
            # No new entities resolved in the last iteration. Exit
            break

    for entity in entities:
        if entity not in resolved:
            resolved.append(entity)

    return resolved


@contextmanager
def recreate_dropped(connection) -> Generator[Session, None, None]:
    """Recreate any dropped all ReplaceableEntities that were dropped within block

    This is useful for making cascading updates. For example, updating a table's column type when it has dependent views.

    def upgrade() -> None:

        my_view = PGView(...)

        with recreate_dropped(op.get_bind()) as conn:

            op.drop_entity(my_view)

            # change an integer column to a bigint
            op.alter_column(
                table_name="account",
                column_name="id",
                schema="public"
                type_=sa.BIGINT()
                existing_type=sa.Integer(),
            )
    """
    from alembic_utils_extended.pg_function import PGFunction
    from alembic_utils_extended.pg_materialized_view import PGMaterializedView
    from alembic_utils_extended.pg_trigger import PGTrigger
    from alembic_utils_extended.pg_view import PGView
    from alembic_utils_extended.replaceable_entity import ReplaceableEntity

    # Do not include permissions here e.g. PGGrantTable. If columns granted to users are dropped, it will cause an error

    def collect_all_db_entities(sess: Session) -> List[ReplaceableEntity]:
        """Collect all entities from the database"""

        return [
            *PGFunction.from_database(sess, "%"),
            *PGTrigger.from_database(sess, "%"),
            *PGView.from_database(sess, "%"),
            *PGMaterializedView.from_database(sess, "%"),
        ]

    sess = Session(bind=connection)

    # All existing entities, before the upgrade
    before = collect_all_db_entities(sess)

    # In the yield, do a
    #     op.drop_entity(my_mat_view, cascade=True)
    #     op.create_entity(my_mat_view)
    try:
        yield sess
    except:
        sess.rollback()
        raise

    # All existing entities, after the upgrade
    after = collect_all_db_entities(sess)
    after_identities = {x.identity for x in after}

    # Entities that were not impacted, or that we have "recovered"
    resolved = []
    unresolved = []

    # First, ignore the ones that were not impacted by the upgrade
    for ent in before:
        if ent.identity in after_identities:
            resolved.append(ent)
        else:
            unresolved.append(ent)

    # Attempt to find an acceptable order of creation for the unresolved entities
    ordered_unresolved = solve_resolution_order(sess, unresolved)

    # Attempt to recreate the missing entities in the specified order
    for ent in ordered_unresolved:
        for stmt in ent.to_sql_statement_create():
            sess.execute(stmt)

    # Sanity check that everything is now fine
    sanity_check = collect_all_db_entities(sess)
    # Fail and rollback if the sanity check is wrong
    try:
        assert len(before) == len(sanity_check)
    except:
        sess.rollback()
        raise

    # Close out the session
    sess.commit()



# pylint: disable=unused-argument,invalid-name,line-too-long
import copy
import logging
from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING, List, Optional

from sqlalchemy.orm import Session

if TYPE_CHECKING:
    from alembic_utils_extended.replaceable_entity import ReplaceableEntity


logger = logging.getLogger(__name__)


@contextmanager
def simulate_entity(
    sess: Session,
    entity: "ReplaceableEntity",
    dependencies: Optional[List["ReplaceableEntity"]] = None,
):
    """Creates *entiity* in a transaction so postgres rendered definition
    can be retrieved
    """

    # When simulating materialized view, don't populate them with data
    from alembic_utils_extended.pg_materialized_view import PGMaterializedView

    if isinstance(entity, PGMaterializedView) and entity.with_data:
        entity = copy.deepcopy(entity)
        entity.with_data = False

    deps: List["ReplaceableEntity"] = dependencies or []

    outer_transaction = sess.begin_nested()
    try:
        dependency_managers = [simulate_entity(sess, x) for x in deps]

        with ExitStack() as stack:
            # Setup all the possible deps
            for mgr in dependency_managers:
                stack.enter_context(mgr)

            did_drop = False
            inner_transaction = sess.begin_nested()
            try:
                for stmt in entity.to_sql_statement_drop(cascade=True):
                    sess.execute(stmt)
                did_drop = True

                for stmt in entity.to_sql_statement_create():
                    sess.execute(stmt)

                yield sess
            except:
                if did_drop:
                    # The drop was successful, so either create was not, or the
                    # error came from user code after the yield.
                    # Anyway, we can exit now.
                    raise

                # Try again without the drop in case the drop raised
                # a does not exist error
                inner_transaction.rollback()
                inner_transaction = sess.begin_nested()

                for stmt in entity.to_sql_statement_create():
                    sess.execute(stmt)

                yield sess
            finally:
                inner_transaction.rollback()
    finally:
        outer_transaction.rollback()



from typing import Generator

from parse import parse
from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import SQLParseFailure
from alembic_utils_extended.on_entity_mixin import OnEntityMixin
from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import coerce_to_quoted


class PGPolicy(OnEntityMixin, ReplaceableEntity):
    """A PostgreSQL Policy compatible with `alembic revision --autogenerate`

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A SQL policy name and tablename, separated by "."
    * **definition** - *str*:  The definition of the policy, incl. permissive, for, to, using, with check
    * **on_entity** - *str*:  fully qualifed entity that the policy applies
    """

    type_ = "policy"

    @classmethod
    def from_sql(cls, sql: str) -> "PGPolicy":
        """Create an instance instance from a SQL string"""

        template = "create policy{:s}{signature}{:s}on{:s}{on_entity}{:s}{definition}"
        result = parse(template, sql.strip(), case_sensitive=False)

        if result is not None:

            on_entity = result["on_entity"]
            if "." not in on_entity:
                schema = "public"
                on_entity = schema + "." + on_entity
            schema, _, _ = on_entity.partition(".")

            return cls(  # type: ignore
                schema=schema,
                signature=result["signature"],
                definition=result["definition"],
                on_entity=on_entity,
            )
        raise SQLParseFailure(f'Failed to parse SQL into PGPolicy """{sql}"""')

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create poicy" statement for PGPolicy"""

        yield sql_text(f"CREATE POLICY {self.signature} on {self.on_entity} {self.definition}")

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop policy" statement for PGPolicy"""
        cascade = "cascade" if cascade else ""
        yield sql_text(f"DROP POLICY {self.signature} on {self.on_entity} {cascade}")

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Not implemented, postgres policies do not support replace."""
        yield sql_text(f"DROP POLICY IF EXISTS {self.signature} on {self.on_entity};")
        yield sql_text(f"CREATE POLICY {self.signature} on {self.on_entity} {self.definition};")

    @classmethod
    def from_database(cls, connection, schema):
        """Get a list of all policies defined in the db"""
        sql = sql_text(
            f"""
        select
            schemaname,
            tablename,
            policyname,
            permissive,
            roles,
            cmd,
            qual,
            with_check
        from
            pg_policies
        where
            schemaname = '{schema}'
        """
        )
        rows = connection.execute(sql).fetchall()

        def get_definition(permissive, roles, cmd, qual, with_check):
            definition = ""
            if permissive is not None:
                definition += f"as {permissive} "
            if cmd is not None:
                definition += f"for {cmd} "
            if roles is not None:
                definition += f"to {', '.join(roles)} "
            if qual is not None:
                if qual[0] != "(":
                    qual = f"({qual})"
                definition += f"using {qual} "
            if with_check is not None:
                if with_check[0] != "(":
                    with_check = f"({with_check})"
                definition += f"with check {with_check} "
            return definition

        db_policies = []
        for schema, table, policy_name, permissive, roles, cmd, qual, with_check in rows:
            definition = get_definition(permissive, roles, cmd, qual, with_check)

            schema = coerce_to_quoted(schema)
            table = coerce_to_quoted(table)
            policy_name = coerce_to_quoted(policy_name)
            policy = cls.from_sql(f"create policy {policy_name} on {schema}.{table} {definition}")
            db_policies.append(policy)

        for policy in db_policies:
            assert policy is not None

        return db_policies



from typing import TYPE_CHECKING, Any, Tuple, Type

from alembic.autogenerate import renderers
from alembic.operations import MigrateOperation, Operations
from typing_extensions import Protocol

from alembic_utils_extended.exceptions import UnreachableException

if TYPE_CHECKING:
    from alembic_utils_extended.replaceable_entity import ReplaceableEntity


class SupportsTarget(Protocol):
    def __init__(self, target: "ReplaceableEntity") -> None:
        pass


class SupportsTargetCascade(Protocol):
    def __init__(self, target: "ReplaceableEntity", cascade: bool) -> None:
        pass


class ReversibleOp(MigrateOperation):
    """A SQL operation that can be reversed"""

    def __init__(self, target: "ReplaceableEntity"):
        self.target = target

    @classmethod
    def invoke_for_target(cls: Type[SupportsTarget], operations, target: "ReplaceableEntity"):
        op = cls(target)
        return operations.invoke(op)

    @classmethod
    def invoke_for_target_optional_cascade(cls: Type[SupportsTargetCascade], operations, target: "ReplaceableEntity", cascade=False):
        op = cls(target, cascade=cascade)  # pylint: disable=unexpected-keyword-arg
        return operations.invoke(op)

    def reverse(self):
        raise NotImplementedError()


##############
# Operations #
##############


@Operations.register_operation("create_entity", "invoke_for_target")
class CreateOp(ReversibleOp):
    def reverse(self):
        return DropOp(self.target)

    def to_diff_tuple(self) -> Tuple[Any, ...]:
        return (
            "create_entity",
            self.target.identity,
            "".join(str(x) for x in self.target.to_sql_statement_create()),
        )


@Operations.register_operation("drop_entity", "invoke_for_target_optional_cascade")
class DropOp(ReversibleOp):
    def __init__(self, target: "ReplaceableEntity", cascade: bool = False) -> None:
        self.cascade = cascade
        super().__init__(target)

    def reverse(self):
        return CreateOp(self.target)

    def to_diff_tuple(self) -> Tuple[Any, ...]:
        return "drop_entity", self.target.identity


@Operations.register_operation("replace_entity", "invoke_for_target")
class ReplaceOp(ReversibleOp):
    def reverse(self):
        return RevertOp(self.target)

    def to_diff_tuple(self) -> Tuple[Any, ...]:
        return (
            "replace_or_revert_entity",
            self.target.identity,
            str([str(x) for x in self.target.to_sql_statement_create_or_replace()]),
        )


class RevertOp(ReversibleOp):
    # Revert is never in an upgrade, so no need to implement reverse

    def to_diff_tuple(self) -> Tuple[Any, ...]:
        return (
            "replace_or_revert_entity",
            self.target.identity,
            str([str(x) for x in self.target.to_sql_statement_create_or_replace()]),
        )


###################
# Implementations #
###################


@Operations.implementation_for(CreateOp)
def create_entity(operations, operation):
    target: "ReplaceableEntity" = operation.target
    for stmt in target.to_sql_statement_create():
        operations.execute(stmt)


@Operations.implementation_for(DropOp)
def drop_entity(operations, operation):
    target: "ReplaceableEntity" = operation.target
    for stmt in target.to_sql_statement_drop(cascade=operation.cascade):
        operations.execute(stmt)


@Operations.implementation_for(ReplaceOp)
@Operations.implementation_for(RevertOp)
def replace_or_revert_entity(operations, operation):
    target: "ReplaceableEntity" = operation.target
    for stmt in target.to_sql_statement_create_or_replace():
        operations.execute(stmt)


##########
# Render #
##########


@renderers.dispatch_for(CreateOp)
def render_create_entity(autogen_context, op):
    target = op.target
    autogen_context.imports.add(target.render_import_statement())
    variable_name = target.to_variable_name()
    return target.render_self_for_migration() + f"op.create_entity({variable_name})\n" + target.render_post_create_entity(autogen_context)


@renderers.dispatch_for(DropOp)
def render_drop_entity(autogen_context, op):
    target = op.target
    autogen_context.imports.add(target.render_import_statement())
    variable_name = target.to_variable_name()
    return target.render_self_for_migration(omit_definition=False) + f"op.drop_entity({variable_name})\n"


@renderers.dispatch_for(ReplaceOp)
def render_replace_entity(autogen_context, op):
    target = op.target
    autogen_context.imports.add(target.render_import_statement())
    variable_name = target.to_variable_name()
    return target.render_self_for_migration() + f"op.replace_entity({variable_name})\n" + target.render_post_create_entity(autogen_context)


@renderers.dispatch_for(RevertOp)
def render_revert_entity(autogen_context, op):
    """Collect the entity definition currently live in the database and use its definition
    as the downgrade revert target"""
    # At the time is call is made, the engine is disconnected

    # We should never reach this call unless an update's revert is being rendered
    # In that case, get_required_migration_op  has cached the database's liver version
    # as target._version_to_replace

    target = op.target
    autogen_context.imports.add(target.render_import_statement())

    db_target = target._version_to_replace

    if db_target is None:
        raise UnreachableException

    variable_name = db_target.to_variable_name()
    return db_target.render_self_for_migration() + f"op.replace_entity({variable_name})"



# pylint: disable=unused-argument,invalid-name,line-too-long


from typing import Generator

from parse import parse
from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import SQLParseFailure
from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import (
    coerce_to_unquoted,
    normalize_whitespace,
    strip_terminating_semicolon,
)


class PGView(ReplaceableEntity):
    """A PostgreSQL View compatible with `alembic revision --autogenerate`

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A SQL view's call signature
    * **definition** - *str*: The SQL select statement body of the view
    """

    type_ = "view"

    def __init__(self, schema: str, signature: str, definition: str):
        self.schema: str = coerce_to_unquoted(normalize_whitespace(schema))
        self.signature: str = coerce_to_unquoted(normalize_whitespace(signature))
        self.definition: str = strip_terminating_semicolon(definition)

    @classmethod
    def from_sql(cls, sql: str) -> "PGView":
        """Create an instance from a SQL string"""
        template = "create{}view{:s}{schema}.{signature}{:s}as{:s}{definition}"
        result = parse(template, sql, case_sensitive=False)
        if result is not None:
            # If the signature includes column e.g. my_view (col1, col2, col3) remove them
            signature = result["signature"].split("(")[0]
            return cls(
                schema=result["schema"],
                # strip quote characters
                signature=signature.replace('"', ""),
                definition=strip_terminating_semicolon(result["definition"]),
            )

        raise SQLParseFailure(f'Failed to parse SQL into PGView """{sql}"""')

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create view" statement"""
        yield sql_text(f'CREATE VIEW {self.literal_schema}."{self.signature}" AS {self.definition};')

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop view" statement"""
        cascade = "cascade" if cascade else ""
        yield sql_text(f'DROP VIEW {self.literal_schema}."{self.signature}" {cascade}')

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create or replace view" statement

        If the initial "CREATE OR REPLACE" statement does not succeed,
        fails over onto "DROP VIEW" followed by "CREATE VIEW"
        """
        yield sql_text(
            f"""
        do $$
            begin
                CREATE OR REPLACE VIEW {self.literal_schema}."{self.signature}" AS {self.definition};

            exception when others then
                DROP VIEW IF EXISTS {self.literal_schema}."{self.signature}";

                CREATE VIEW {self.literal_schema}."{self.signature}" AS {self.definition};
            end;
        $$ language 'plpgsql'
        """
        )

    @classmethod
    def from_database(cls, sess, schema):
        """Get a list of all functions defined in the db"""
        sql = sql_text(
            f"""
        select
            schemaname schema_name,
            viewname view_name,
            definition
        from
            pg_views
        where
            schemaname not in ('pg_catalog', 'information_schema')
            and schemaname::text like '{schema}';
        """
        )
        rows = sess.execute(sql).fetchall()
        db_views = [cls(x[0], x[1], x[2]) for x in rows]

        for view in db_views:
            assert view is not None

        return db_views



# pylint: disable=unused-argument,invalid-name,line-too-long
import logging
from itertools import zip_longest
from pathlib import Path
from typing import (
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

from alembic.autogenerate import comparators
from alembic.autogenerate.api import AutogenContext
from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import TextClause

import alembic_utils_extended
from alembic_utils_extended.depends import solve_resolution_order
from alembic_utils_extended.exceptions import UnreachableException
from alembic_utils_extended.experimental import collect_subclasses
from alembic_utils_extended.reversible_op import (
    CreateOp,
    DropOp,
    ReplaceOp,
    ReversibleOp,
)
from alembic_utils_extended.simulate import simulate_entity
from alembic_utils_extended.statement import (
    coerce_to_quoted,
    coerce_to_unquoted,
    escape_colon_for_sql,
    normalize_whitespace,
    strip_terminating_semicolon,
)

logger = logging.getLogger(__name__)

T = TypeVar("T", bound="ReplaceableEntity")


class ReplaceableEntity:
    """A SQL Entity that can be replaced"""

    def __init__(self, schema: str, signature: str, definition: str):
        self.schema: str = coerce_to_unquoted(normalize_whitespace(schema))
        self.signature: str = coerce_to_unquoted(normalize_whitespace(signature))
        self.definition: str = escape_colon_for_sql(strip_terminating_semicolon(definition))

    @property
    def type_(self) -> str:
        """In order to support calls to `run_name_filters` and
        `run_object_filters` on the AutogenContext object, each
        entity needs to have a named type.
        https://alembic.sqlalchemy.org/en/latest/api/autogenerate.html#alembic.autogenerate.api.AutogenContext.run_name_filters
        """
        raise NotImplementedError()

    @classmethod
    def from_sql(cls: Type[T], sql: str) -> T:
        """Create an instance from a SQL string"""
        raise NotImplementedError()

    @property
    def literal_schema(self) -> str:
        """Wrap a schema name in literal quotes
        Useful for emitting SQL statements
        """
        return coerce_to_quoted(self.schema)

    @classmethod
    def from_path(cls: Type[T], path: Path) -> T:
        """Create an instance instance from a SQL file path"""
        with path.open() as sql_file:
            sql = sql_file.read()
        return cls.from_sql(sql)

    @classmethod
    def from_database(cls, sess: Session, schema="%") -> List[T]:
        """Collect existing entities from the database for given schema"""
        raise NotImplementedError()

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create function" statement for PGFunction"""
        raise NotImplementedError()

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop function" statement for PGFunction"""
        raise NotImplementedError()

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create or replace function" statement for PGFunction"""
        raise NotImplementedError()

    def get_database_definition(self: T, sess: Session, dependencies: Optional[List["ReplaceableEntity"]] = None) -> T:  # $Optional[T]:
        """Creates the entity in the database, retrieves its 'rendered' then rolls it back"""
        with simulate_entity(sess, self, dependencies) as sess:
            # Drop self
            for stmt in self.to_sql_statement_drop():
                sess.execute(stmt)

            # collect all remaining entities
            db_entities: List[T] = sorted(self.from_database(sess, schema=self.schema), key=lambda x: x.identity)

        with simulate_entity(sess, self, dependencies) as sess:
            # collect all remaining entities
            all_w_self: List[T] = sorted(self.from_database(sess, schema=self.schema), key=lambda x: x.identity)

        # Find "self" by diffing the before and after
        for without_self, with_self in zip_longest(db_entities, all_w_self):
            if without_self is None or without_self.identity != with_self.identity:
                return with_self

        raise UnreachableException()

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__
        escaped_definition = self.definition if not omit_definition else "# not required for op"

        return f"""{var_name} = {class_name}(
    schema="{self.schema}",
    signature="{self.signature}",
    definition={repr(escaped_definition)}
)\n"""

    # Default to no additional operations.
    def render_post_create_entity(self, _autogen_context: AutogenContext) -> str:
        return ""

    @classmethod
    def render_import_statement(cls) -> str:
        """Render a string that is valid python code to import current class"""
        module_path = cls.__module__
        class_name = cls.__name__
        return f"from {module_path} import {class_name}\nfrom sqlalchemy import text as sql_text"

    @property
    def identity(self) -> str:
        """A string that consistently and globally identifies a function"""
        return f"{self.__class__.__name__}: {self.schema}.{self.signature}"

    def to_variable_name(self) -> str:
        """A deterministic variable name based on PGFunction's contents"""
        schema_name = self.schema.lower()
        object_name = self.signature.split("(")[0].strip().lower().replace("-", "_")
        return f"{schema_name}_{object_name}"

    _version_to_replace: Optional[T] = None  # type: ignore

    def get_required_migration_op(
        self: T, sess: Session, dependencies: Optional[List["ReplaceableEntity"]] = None
    ) -> Optional[ReversibleOp]:
        """Get the migration operation required for autogenerate"""
        # All entities in the database for self's schema
        entities_in_database: List[T] = self.from_database(sess, schema=self.schema)

        db_def = self.get_database_definition(sess, dependencies=dependencies)

        for x in entities_in_database:

            if (db_def.identity, normalize_whitespace(db_def.definition)) == (
                x.identity,
                normalize_whitespace(x.definition),
            ):
                return None

            if db_def.identity == x.identity:
                # Cache the currently live copy to render a RevertOp without hitting DB again
                self._version_to_replace = x
                return ReplaceOp(self)

        return CreateOp(self)


class ReplaceableEntityRegistry:
    def __init__(self):
        self._entities: Dict[str, ReplaceableEntity] = {}
        self.schemas: Set[str] = set()
        self.exclude_schemas: Set[str] = set()
        self.entity_types: Set[Type[ReplaceableEntity]] = set()

    def clear(self):
        self._entities.clear()
        self.schemas.clear()
        self.exclude_schemas.clear()
        self.entity_types.clear()

    def register(
        self,
        entities: Iterable[ReplaceableEntity],
        schemas: Optional[List[str]] = None,
        exclude_schemas: Optional[Iterable[str]] = None,
        entity_types: Optional[Iterable[Type[ReplaceableEntity]]] = None,
    ) -> None:
        self._entities.update({e.identity: e for e in entities})

        if schemas:
            self.schemas |= set(schemas)

        if exclude_schemas:
            self.exclude_schemas |= set(exclude_schemas)

        if entity_types:
            self.entity_types |= set(entity_types)

    @property
    def allowed_entity_types(self) -> Set[Type[ReplaceableEntity]]:
        if self.entity_types:
            return self.entity_types
        return set(collect_subclasses(alembic_utils_extended, ReplaceableEntity))

    def entities(self) -> List[ReplaceableEntity]:
        return list(self._entities.values())


registry = ReplaceableEntityRegistry()


##################
# Event Listener #
##################


def register_entities(
    entities: Iterable[ReplaceableEntity],
    schemas: Optional[List[str]] = None,
    exclude_schemas: Optional[Iterable[str]] = None,
    entity_types: Optional[Iterable[Type[ReplaceableEntity]]] = None,
) -> None:
    """Register entities to be monitored for changes when alembic is invoked with `revision --autogenerate`.

    **Parameters:**

    * **entities** - *List[ReplaceableEntity]*: A list of entities (PGFunction, PGView, etc) to monitor for revisions

    **Deprecated Parameters:**

    *Configure schema and object inclusion/exclusion with `include_name` and `include_object` in `env.py`. For more information see https://alembic.sqlalchemy.org/en/latest/autogenerate.html#controlling-what-to-be-autogenerated*


    * **schemas** - *Optional[List[str]]*: A list of SQL schema names to monitor. Note, schemas referenced in registered entities are automatically monitored.
    * **exclude_schemas** - *Optional[List[str]]*: A list of SQL schemas to ignore. Note, explicitly registered entities will still be monitored.
    * **entity_types** - *Optional[List[Type[ReplaceableEntity]]]*: A list of ReplaceableEntity classes to consider during migrations. Other entity types are ignored
    """
    registry.register(entities, schemas, exclude_schemas, entity_types)


@comparators.dispatch_for("schema")
def compare_registered_entities(
    autogen_context: AutogenContext,
    upgrade_ops,
    schemas: List[Optional[str]],
):
    connection = autogen_context.connection

    entities = registry.entities()

    # https://alembic.sqlalchemy.org/en/latest/autogenerate.html#controlling-what-to-be-autogenerated
    # if EnvironmentContext.configure.include_schemas is True, all non-default "scehmas" should be included
    # pulled from the inspector
    include_schemas: bool = autogen_context.opts["include_schemas"]

    reflected_schemas = set(autogen_context.inspector.get_schema_names() if include_schemas else [])  # type: ignore
    sqla_schemas: Set[Optional[str]] = set(schemas)
    manual_schemas = set(registry.schemas or set())  # Deprecated for remove in 0.6.0
    entity_schemas = {x.schema for x in entities}  # from ReplaceableEntity instances
    all_schema_references = reflected_schemas | sqla_schemas | manual_schemas | entity_schemas  # type: ignore

    # Remove excluded schemas
    observed_schemas: Set[str] = {
        schema_name
        for schema_name in all_schema_references
        if (
            schema_name is not None not in (registry.exclude_schemas or set())  # user defined. Deprecated for remove in 0.6.0
            and schema_name not in {"information_schema", None}
        )
    }

    # Solve resolution order
    transaction = connection.begin_nested()
    sess = Session(bind=connection)
    try:
        ordered_entities: List[ReplaceableEntity] = solve_resolution_order(sess, entities)
    finally:
        sess.rollback()

    # entities that are receiving a create or update op
    has_create_or_update_op: List[ReplaceableEntity] = []

    # database rendered definitions for the entities we have a local instance for
    # Note: used for drops
    local_entities = []

    # Required migration OPs, Create/Update/NoOp
    for entity in ordered_entities:
        logger.info(
            "Detecting required migration op %s %s",
            entity.__class__.__name__,
            entity.identity,
        )

        if entity.__class__ not in registry.allowed_entity_types:
            continue

        if not include_entity(entity, autogen_context, reflected=False):
            logger.debug(
                "Ignoring local entity %s %s due to AutogenContext filters",
                entity.__class__.__name__,
                entity.identity,
            )
            continue

        transaction = connection.begin_nested()
        sess = Session(bind=connection)
        try:
            maybe_op = entity.get_required_migration_op(sess, dependencies=has_create_or_update_op)

            local_db_def = entity.get_database_definition(sess, dependencies=has_create_or_update_op)
            local_entities.append(local_db_def)

            if maybe_op:
                upgrade_ops.ops.append(maybe_op)
                has_create_or_update_op.append(entity)

                logger.info(
                    "Detected %s op for %s %s",
                    maybe_op.__class__.__name__,
                    entity.__class__.__name__,
                    entity.identity,
                )
            else:
                logger.debug(
                    "Detected NoOp op for %s %s",
                    entity.__class__.__name__,
                    entity.identity,
                )

        finally:
            sess.rollback()

    # Required migration OPs, Drop
    # Start a parent transaction
    # Bind the session within the parent transaction
    transaction = connection.begin_nested()
    sess = Session(bind=connection)
    try:
        # All database entities currently live
        # Check if anything needs to drop
        subclasses = collect_subclasses(alembic_utils_extended, ReplaceableEntity)
        for entity_class in subclasses:

            if entity_class not in registry.allowed_entity_types:
                continue

            # Entities within the schemas that are live
            for schema in observed_schemas:

                db_entities: List[ReplaceableEntity] = entity_class.from_database(sess, schema=schema)

                # Check for functions that were deleted locally
                for db_entity in db_entities:

                    if not include_entity(db_entity, autogen_context, reflected=True):
                        logger.debug(
                            "Ignoring remote entity %s %s due to AutogenContext filters",
                            db_entity.__class__.__name__,
                            db_entity.identity,
                        )
                        continue

                    for local_entity in local_entities:
                        if db_entity.identity == local_entity.identity:
                            break
                    else:
                        # No match was found locally
                        # If the entity passes the filters,
                        # we should create a DropOp
                        upgrade_ops.ops.append(DropOp(db_entity))
                        logger.info(
                            "Detected DropOp op for %s %s",
                            db_entity.__class__.__name__,
                            db_entity.identity,
                        )

    finally:
        sess.rollback()


def include_entity(entity: ReplaceableEntity, autogen_context: AutogenContext, reflected: bool) -> bool:
    """The functions on the AutogenContext object
    are described here:
    https://alembic.sqlalchemy.org/en/latest/api/autogenerate.html#alembic.autogenerate.api.AutogenContext.run_name_filters

    The meaning of the function parameters are explained in the corresponding
    definitions in the EnvironmentContext object:
    https://alembic.sqlalchemy.org/en/latest/api/runtime.html#alembic.runtime.environment.EnvironmentContext.configure.params.include_name

    This will only have an impact for projects which set include_object and/or include_name in the configuration
    of their Alembic env.
    """
    name = f"{entity.schema}.{entity.signature}"
    parent_names = {
        "schema_name": entity.schema,
        # At the time of writing, the implementation of `run_name_filters` in Alembic assumes that every type of object
        # will either be a table or have a table_name in its `parent_names` dict. This is true for columns and indexes,
        # but not true for the type of objects supported in this library such as views as functions. Nevertheless, to avoid
        # a KeyError when calling `run_name_filters`, we have to set some value.
        "table_name": f"Not applicable for type {entity.type_}",
    }
    # According to the Alembic docs, the name filter is only relevant for reflected objects
    if reflected:
        name_result = autogen_context.run_name_filters(name, entity.type_, parent_names)
    else:
        name_result = True

    # Object filters should be applied to object from local metadata and to reflected objects
    object_result = autogen_context.run_object_filters(entity, name, entity.type_, reflected=reflected, compare_to=None)
    return name_result and object_result



from typing import TYPE_CHECKING

from alembic_utils_extended.statement import coerce_to_unquoted

if TYPE_CHECKING:
    from alembic_utils_extended.replaceable_entity import ReplaceableEntity

    _Base = ReplaceableEntity
else:
    _Base = object


class OnEntityMixin(_Base):
    """Mixin to ReplaceableEntity providing setup for entity types requiring an "ON" clause"""

    def __init__(self, schema: str, signature: str, definition: str, on_entity: str):
        super().__init__(schema=schema, signature=signature, definition=definition)

        if "." not in on_entity:
            on_entity = "public." + on_entity

        # Guarenteed to have a schema
        self.on_entity = coerce_to_unquoted(on_entity)

    @property
    def identity(self) -> str:
        """A string that consistently and globally identifies a function

        Overriding default to add the "on table" clause
        """
        return f"{self.__class__.__name__}: {self.schema}.{self.signature} {self.on_entity}"

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__
        escaped_definition = self.definition if not omit_definition else "# not required for op"

        return f"""{var_name} = {class_name}(
    schema="{self.schema}",
    signature="{self.signature}",
    on_entity="{self.on_entity}",
    definition={repr(escaped_definition)}
)\n"""

    def to_variable_name(self) -> str:
        """A deterministic variable name based on PGFunction's contents"""
        schema_name = self.schema.lower()
        object_name = self.signature.split("(")[0].strip().lower()
        _, _, unqualified_entity_name = self.on_entity.lower().partition(".")
        return f"{schema_name}_{unqualified_entity_name}_{object_name}"



# pylint: disable=unused-argument,invalid-name,line-too-long
from typing import Generator

from parse import parse
from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import SQLParseFailure
from alembic_utils_extended.on_entity_mixin import OnEntityMixin
from alembic_utils_extended.replaceable_entity import ReplaceableEntity


class PGTrigger(OnEntityMixin, ReplaceableEntity):
    """A PostgreSQL Trigger compatible with `alembic revision --autogenerate`

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A SQL trigger's call signature
    * **definition** - *str*:  The remainig trigger body and identifiers
    * **on_entity** - *str*:  fully qualifed entity that the policy applies
    * **is_constraint** - *bool*: Is the trigger a constraint trigger

    Postgres Create Trigger Specification:

        CREATE [ CONSTRAINT ] TRIGGER name { BEFORE | AFTER | INSTEAD OF } { event [ OR ... ] }
        ON table
        [ FROM referenced_table_name ]
        [ NOT DEFERRABLE | [ DEFERRABLE ] { INITIALLY IMMEDIATE | INITIALLY DEFERRED } ]
        [ FOR [ EACH ] { ROW | STATEMENT } ]
        [ WHEN ( condition ) ]
        EXECUTE PROCEDURE function_name ( arguments )
    """

    type_ = "trigger"

    _templates = [
        "create{:s}constraint{:s}trigger{:s}{signature}{:s}{event}{:s}ON{:s}{on_entity}{:s}{action}",
        "create{:s}trigger{:s}{signature}{:s}{event}{:s}ON{:s}{on_entity}{:s}{action}",
    ]

    def __init__(
        self,
        schema: str,
        signature: str,
        definition: str,
        on_entity: str,
        is_constraint: bool = False,
    ):
        super().__init__(schema=schema, signature=signature, definition=definition, on_entity=on_entity)  # type: ignore
        self.is_constraint = is_constraint

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__
        escaped_definition = self.definition if not omit_definition else "# not required for op"

        return f"""{var_name} = {class_name}(
    schema="{self.schema}",
    signature="{self.signature}",
    on_entity="{self.on_entity}",
    is_constraint={self.is_constraint},
    definition={repr(escaped_definition)}
)\n"""

    @property
    def identity(self) -> str:
        """A string that consistently and globally identifies a trigger"""
        return f"{self.__class__.__name__}: {self.schema}.{self.signature} {self.is_constraint} {self.on_entity}"

    @classmethod
    def from_sql(cls, sql: str) -> "PGTrigger":
        """Create an instance instance from a SQL string"""
        for template in cls._templates:
            result = parse(template, sql, case_sensitive=False)
            if result is not None:
                # remove possible quotes from signature
                signature = result["signature"]
                event = result["event"]
                on_entity = result["on_entity"]
                action = result["action"]
                is_constraint = "constraint" in template

                if "." not in on_entity:
                    on_entity = "public" + "." + on_entity

                schema = on_entity.split(".")[0]

                definition_template = " {event} ON {on_entity} {action}"
                definition = definition_template.format(event=event, on_entity=on_entity, action=action)

                return cls(
                    schema=schema,
                    signature=signature,
                    on_entity=on_entity,
                    definition=definition,
                    is_constraint=is_constraint,
                )
        raise SQLParseFailure(f'Failed to parse SQL into PGTrigger """{sql}"""')

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create trigger" statement for PGTrigger"""

        # We need to parse and replace the schema qualifier on the table for simulate_entity to
        # operate
        _def = self.definition
        _template = "{event}{:s}ON{:s}{on_entity}{:s}{action}"
        match = parse(_template, _def)
        if not match:
            raise SQLParseFailure(f'Failed to parse SQL into PGTrigger.definition """{_def}"""')

        event = match["event"]
        action = match["action"]

        # Ensure entity is qualified with schema
        on_entity = match["on_entity"]
        if "." in on_entity:
            _, _, on_entity = on_entity.partition(".")
        on_entity = f"{self.schema}.{on_entity}"

        # Re-render the definition ensuring the table is qualified with
        def_rendered = _template.replace("{:s}", " ").format(event=event, on_entity=on_entity, action=action)

        yield sql_text(f"CREATE{' CONSTRAINT ' if self.is_constraint else ' '}TRIGGER \"{self.signature}\" {def_rendered}")

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop trigger" statement for PGTrigger"""
        cascade = "cascade" if cascade else ""
        yield sql_text(f'DROP TRIGGER "{self.signature}" ON {self.on_entity} {cascade}')

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "replace trigger" statement for PGTrigger"""
        yield sql_text(f'DROP TRIGGER IF EXISTS "{self.signature}" ON {self.on_entity};')
        yield from self.to_sql_statement_create()

    @classmethod
    def from_database(cls, sess, schema):
        """Get a list of all triggers defined in the db"""

        sql = sql_text(
            """
        select
            pc.relnamespace::regnamespace::text as table_schema,
            tgname trigger_name,
            pg_get_triggerdef(pgt.oid) definition
        from
            pg_trigger pgt
                inner join pg_class pc
                    on pgt.tgrelid = pc.oid
        where
            not tgisinternal
            and pc.relnamespace::regnamespace::text like :schema
        """
        )
        rows = sess.execute(sql, {"schema": schema}).fetchall()

        db_triggers = [cls.from_sql(x[2]) for x in rows]

        for trig in db_triggers:
            assert trig is not None

        return db_triggers
